﻿using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace SortedConvexPolygon
{

    // Classe che implementa IComparer, di cui ci serviremo per
    // ordinare i vertici in senso orario
    public class ClockwiseComparer : IComparer<Vector2>
    {
        // Variabile che conterrà il centro del poligono
        private Vector2 center;
        public Vector2 VertexCenter
        {
            get { return center; }
        }
        // Primo punto del poligono che useremo come riferimento
        // rispetto al quale calcolare tutti gli angoli per poi
        // poter ordinare.
        private Vector2 firstPoint;
        // Lunghezza del primo segmento, ovvero distanza del 
        // centro dal primo punto
        private float firstSegmentLength;

        public ClockwiseComparer(IList<Vector2> ptVertexes)
        {
            // Calcoliamo il centro come media di tutti i punti
            foreach (Vector2 pt in ptVertexes)
            {
                center.X += pt.X / ptVertexes.Count;
                center.Y += pt.Y / ptVertexes.Count;
            }

            // Memorizziamo il primo punto, che useremo come riferimento
            firstPoint = ptVertexes[0];

            // Calcoliamo la distanza del primo punto dal centro, ci servirà
            // per poter applicare il teorema del coseno (o teorema di Carnot)
            this.firstSegmentLength = Distance(center, firstPoint);
        }

        // Funzione statica che calcola la distanza tra due vettori utilizzando
        // il teorema di pitagora
        static private float Distance(Vector2 pt1, Vector2 pt2)
        {
            return (float)Math.Sqrt(Math.Pow((pt1.X - pt2.X), 2) + Math.Pow((pt1.Y - pt2.Y), 2));
        }

        // Implementiamo la funzione di confronto che servirà per ordinare la lista
        // di vertici
        public int Compare(Vector2 x, Vector2 y)
        {
            // Un vertice viene prima di un altro in senso orario
            // se l'angolo che forma con il segmento di riferimento
            // è minore di quello dell'altro punto
            return GetAngle(x).CompareTo(GetAngle(y));
        }

        // Funzione che calcola un valore proporzionale all'angolo
        // compreso tra il segmento che unisce il centro e il punto
        // pt e il segmento di riferimento (quello che congiunge) il
        // centro con il primo punto del poligono.
        // Il valore restituito non è un angolo ma un valore ad esso
        // proporzionale compreso tra -2 e +2, ottenuto dal teorema
        // del coseno (o di Carnot).
        public float GetAngle(Vector2 pt)
        {
            // Teorema di Carnot:
            // a^2 = b^2 + c^2 - 2 * b * c * cos(alpha)
            //              b^2 + c^2 - a^2
            // cos(alpha) = ---------------
            //                 2 * b * c

            // Prepariamo le variabili che servono per il teorema
            float a, b, c;
            // Distanza del primo punto del poligono dal punto corrente
            a = Distance(firstPoint, pt);
            // Distanza del primo punto del poligono dal centro
            b = this.firstSegmentLength;
            // Distanza del punto corrente dal centro
            c = Distance(pt, center);

            // Applichiamo il teorema del coseno
            float cos =
                (float)(Math.Pow(b, 2) + Math.Pow(c, 2) - Math.Pow(a, 2)) /
                (2 * b * c);

            // cos è compreso tra -1 e +1, aggiungiamo 1 in modo che sia
            // compreso tra 0 e +2
            cos++;

            // Decidiamo il segno in base alla posizione rispetto al segmento
            // di riferimento, a seconda che siamo prima o dopo in senso orario.
            // Per fare questo vediamo se l'angolo compreso tra il punto corrente,
            // il primo punto del poligono e il centro è positivo o negativo, tramite
            // il prodotto vettore (cross-product)
            if (SortedConvexPolygonGame.CrossProduct(pt - firstPoint, center - firstPoint) > 0)
                cos *= -1;
            return cos;
        }
    }
}
