using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace SortedConvexPolygon
{
    public class SortedConvexPolygonGame : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        // Lo SpriteBatch con cui disegneremo
        SpriteBatch spriteBatch;
        // Una texture che useremo per disegnare i punti
        Texture2D redSquare;
        // Il font che utilizzeremo per scrivere
        SpriteFont arialFont;
        // Lista dei vertici del poligono
        private List<Vector2> vertexes = new List<Vector2>();
        // Memorizziamo lo stato del mouse di volta in volta per
        // poter rilevare l'evento di rilascio di un tasto
        private MouseState lastMouseState;
        // Punto che conterr le coordinate del punto che si
        // intende verificare se sia all'interno del poligono.
        // Per cambiarlo fare click destro nella finestra del gioco.
        private Vector2 checkPoint = new Vector2();
        // Memorizziamo se il checkPoint  all'interno del poligono
        private bool isCheckPointInside = false;


        public SortedConvexPolygonGame()
        {
            graphics = new GraphicsDeviceManager(this);

            // Mostriamo il mouse, che ci servir per indicare i 
            // vertici del poligono e il punto di test
            this.IsMouseVisible = true;

            Content.RootDirectory = "Content";
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // Inizializziamo il font con cui scriveremo alcune informazioni
            arialFont = Content.Load<SpriteFont>("Arial");

            // Creiamo una nuova Texture2D di dimensione 1x1 e impostiamo
            // il suo unico pixel su rosso trasparente al 50%
            redSquare = new Texture2D(this.GraphicsDevice, 1, 1);
            redSquare.SetData<Color>(new Color[] { new Color(Color.Red, 128) });
        }

        // Struttura che contiene informazioni su un lato del poligono
        public struct Side
        {
            public float Start;
            public float End;
            public float XStep;
            public float StartX;

            // Costruiamo un lato a partire dai suoi vertici
            public Side(Vector2 pt1, Vector2 pt2)
            {
                // Memorizziamo dove inizia e dove finisce sull'asse Y
                Start = pt1.Y;
                End = pt2.Y;

                // Calcoliamo per ogni unit lungo le Y di cui ci muoviamo
                // quanto e in che direzione ci muoviamo sulle X (in sostanza
                // il coefficiente angolare della retta).
                XStep = (pt2.X - pt1.X) / (pt2.Y - pt1.Y);
                // Memorizziamo le coordinata X del punto pi in alto, da cui
                // partiremo
                if (pt1.Y < pt2.Y) StartX = pt1.X; else StartX = pt2.X;
            }
        }


        // Disegniamo il poligono ciclando tutti i suoi punti e nessuno di pi.
        // Questa funzione prevede che i vertici siano forniti in senso orario.
        private void DrawPolygon()
        {
            // Abbiamo bisogno di almeno tre vertici per poter disegnare
            // un poligono
            if (vertexes.Count < 3) return;

            // Lista dei lati
            LinkedList<Side> sides = new LinkedList<Side>();
            // Nodo della lista corrente su cui lavoreremo tra poco
            LinkedListNode<Side> currentNode = null;
            // Nodo della lista che delimita correntemente il poligono
            // a sinistra/destra
            LinkedListNode<Side> leftBoundNode = null;
            LinkedListNode<Side> rightBoundNode = null;
            // Coordinata Y del limite superiore del poligono.
            // Si tenga presente che l'asse Y procede dall'angolo
            // in alto a sinistra verso il basso.
            float topY = float.MaxValue;
            // Coordinata Y del limite inferiore del poligono
            float bottomY = 0;

            // Creiamo una linked-list di lati, individuiamo i due lati pi in alto
            // che saranno i primi leftBoundNode e rightBoundNode e memorizziamo
            // le coordinate Y degli estremi del poligono (topY e bottomY)
            for (int C1 = 0; C1 < vertexes.Count; C1++)
            {
                // Indica il vertice successivo, ovvero quello con indice 
                // successivo o il primo della lista se quello corrente 
                // l'ultimo, grazie all'operatore modulo.
                int nextIndex = (C1 + 1) % vertexes.Count;
                // Ignoriamo i lati orizzontali
                if (vertexes[C1].Y != vertexes[nextIndex].Y)
                {
                    // Aggiungiamo in fondo alla linked-list il lato corrente
                    // formato dal vertice corrente e dal successivo.
                    currentNode = sides.AddLast(new Side(vertexes[C1], vertexes[nextIndex]));

                    // Individuiamo il limite inferiore del poligono
                    if (vertexes[C1].Y > bottomY) bottomY = vertexes[C1].Y;

                    // Individuiamo il lato destro e sinistro pi in alto
                    // e al contempo troviamo la coordinata Y del limite
                    // superiore del poligono

                    // Individuiamo il lato che inizia pi in alto
                    // che sar quindi il primo limite destro del poligono.
                    if (currentNode.Value.Start <= topY)
                    {
                        topY = currentNode.Value.Start;
                        rightBoundNode = currentNode;
                    }
                    // Individuiamo il lato che finisce pi in alto
                    // che sar quindi il primo limite sinistro del poligono.
                    if (currentNode.Value.End <= topY)
                    {
                        topY = currentNode.Value.End;
                        leftBoundNode = currentNode;
                    }
                }
            }

            // Cursori che useremo per muoverci all'interno del poligono
            // sui due assi.
            float CY = topY;
            float CX = 0;
            // Limiti destro e sinistro entro il quale il cursore
            // lungo l'asse X dovr muoversi.
            float xRightLimit = rightBoundNode.Value.StartX;
            float xLeftLimit = leftBoundNode.Value.StartX;

            // Effettuiamo una scansione dal limite superiore
            // a quello inferiore del poligono
            while (CY < bottomY)
            {
                // Posizioniamo il cursore X sul limite sinistro
                CX = xLeftLimit;
                // Ci muoviamo lungo l'asse X fino al limite destro
                while (CX <= xRightLimit)
                {
                    // Disegniamo il pixel corrente
                    spriteBatch.Draw(redSquare, new Rectangle((int)CX, (int)CY, 1, 1), Color.White);
                    // Avanziamo lungo l'asse X
                    CX += 1;
                }

                // Avanziamo lungo l'asse Y
                CY += 1;

                // Spostiamo i limiti sinistro e destro di un passo
                // a destra o a sinistra a seconda dell'inclinazione
                // del lato corrente.
                xLeftLimit += leftBoundNode.Value.XStep;
                xRightLimit += rightBoundNode.Value.XStep;

                // Se la coordinata Y a cui siamo giunti  al di fuori
                // del range di valori che copre il segmento destro
                // corrente passiamo al successivo
                if (CY > rightBoundNode.Value.End)
                {
                    rightBoundNode = rightBoundNode.Next;
                    // Se siamo giunti all'ultimo lato, passiamo al primo della lista
                    if (rightBoundNode == null) rightBoundNode = sides.First;
                    xRightLimit = rightBoundNode.Value.StartX;
                }
                // Come sopra, ma tutto  invertito per il lato sinistro
                if (CY > leftBoundNode.Value.Start)
                {
                    leftBoundNode = leftBoundNode.Previous;
                    if (leftBoundNode == null) leftBoundNode = sides.Last;
                    xLeftLimit = leftBoundNode.Value.StartX;
                }
            }

        }

        protected override void Update(GameTime gameTime)
        {
            KeyboardState kbd = Keyboard.GetState();

            // Esce dal gioco
            if (kbd.IsKeyDown(Keys.Escape))
                this.Exit();

            // Pulisce la lista dei vertici
            if (kbd.IsKeyDown(Keys.Delete))
                this.vertexes.Clear();

            MouseState ms = Mouse.GetState();

            // Se nel precedente Update il tasto sinistro del mouse era premuto e ora non lo  pi
            if (lastMouseState.LeftButton == ButtonState.Pressed && ms.LeftButton == ButtonState.Released)
            {
                // Punto in cui si  fatto click
                Vector2 ptCurrent = new Vector2(ms.X, ms.Y);

                // Se non  gi presente tra i vertici
                if (!vertexes.Contains(ptCurrent))
                {
                    // Aggiungiamo il punto cliccato alla lista dei vertici
                    vertexes.Add(ptCurrent);
                    // Creiamo un ClockwiseComparer tramite il quale ri-ordiniamo
                    // la lista
                    ClockwiseComparer cmpClockwise = new ClockwiseComparer(vertexes);
                    vertexes.Sort(cmpClockwise);
                }
            }

            // Se nel precedente Update il tasto destro del mouse era premuto e ora non lo  pi
            if (lastMouseState.RightButton == ButtonState.Pressed && ms.RightButton == ButtonState.Released)
            {
                // Aggiorniamo il checkPoint
                checkPoint = new Vector2(ms.X, ms.Y);
                // Verifichiamo se  all'interno del poligono o no
                isCheckPointInside = SortedConvexPolygonGame.IsInsideConvexPolygon(vertexes, checkPoint);
            }

            // Memorizziamo lo stato del mouse corrente
            lastMouseState = ms;
            base.Update(gameTime);
        }

        // Funzione che verifica se un punto  all'interno di un poligono.
        // Per verificarlo in sostanza scansioniamo tutti i lati in senso orario
        // o antiorario e verifichiamo il punto si trovi sempre dalla stessa parte
        // (a destra in senso orario, a sinistra in senso anti-orario) rispetto ad
        // ogni lato.
        static bool IsInsideConvexPolygon(IList<Vector2> vertexes, Vector2 checkPoint)
        {
            // Variabile in cui, dopo il primo ciclo, memorizzeremo
            // se il punto si trova a destra o a sinsitra del lato
            // cos da poter verificare che sia cos per ogni lato.
            sbyte sign = 0;
            // Cicliamo lungo tutti i vertici
            for (int C1 = 0; C1 < vertexes.Count; C1++)
            {
                // Prendiamo il punto corrente e il successivo
                Vector2 pt1 = vertexes[C1];
                Vector2 pt2 = vertexes[(C1 + 1) % vertexes.Count];
                // Sottraiamo (traslazione) al checkPoint e 
                // e ad un vertice del lato corrente l'altro vertice
                // cos da poter calcolare il segno dell'angolo 
                // compreso tra il lato e il checkPoint tramite il
                // prodotto vettore (cross-product)
                Vector2 translatedCheckPoint = checkPoint - pt1;
                pt2 -= pt1;
                // Prendiamo il segno dell'angolo ottenuto tramite il 
                // prodotto vettore.
                sbyte angleSign = (sbyte)Math.Sign(CrossProduct(pt2, translatedCheckPoint));

                // Se  il primo ciclo
                if (sign == 0)
                {
                    sign = angleSign;
                }
                else if (angleSign != sign)
                {
                    // Se non  il primo ciclo e il segno dell'angolo
                    //  differente dagli altri, siamo fuori dal poligono
                    return false;
                }
            }
            // Abbiamo controllato tutti i lati, siamo all'interno del poligono
            return true;
        }

        // Prodotto vettore (cross-product) a due dimensioni
        public static float CrossProduct(Vector2 pt1, Vector2 pt2)
        {
            return pt1.X * pt2.Y - pt1.Y * pt2.X;
        }

        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            spriteBatch.Begin();
            // Disegniamo il poligono
            DrawPolygon();
            // Disegniamo i vertici
            for (int CVertex = 0; CVertex < vertexes.Count; CVertex++)
            {
                spriteBatch.DrawString(arialFont, CVertex.ToString(), vertexes[CVertex] + new Vector2(3, 3), Color.Green);
                spriteBatch.Draw(redSquare, new Rectangle((int)vertexes[CVertex].X, (int)vertexes[CVertex].Y, 5, 5), Color.White);
            }

            // Disegniamo il checkPoint
            spriteBatch.Draw(redSquare, new Rectangle((int)checkPoint.X, (int)checkPoint.Y, 5, 5), Color.White);
            spriteBatch.DrawString(arialFont, isCheckPointInside ? "in" : "out", new Vector2(checkPoint.X, checkPoint.Y) + new Vector2(3, 3), Color.Green);
            spriteBatch.End();
            base.Draw(gameTime);
        }
    }
}
